#!/usr/bin/env python3
import sys
sys.path.append('automation/trex_control_plane/interactive/trex/examples/stl')
STL_PROFILES_PATH = 'stl/'
import stl_path
from trex.stl.api import *

import time
import json
from pprint import pprint
import argparse

import os

def fmt_xps(xps):
    if xps>1e9:
        return ("%.2f G" % (xps / 1e9))
    if xps>1e6:
        return ("%.2f M" % (xps / 1e6))
    if xps>1e3:
        return ("%.2f K" % (xps / 1e3))
    return ("%.0f " % xps)

##### Load Test
# Given a server structure and an input dictionary:
# args['profile_file']    = "udp_1pkt.py"     # Profile to use (eg. "udp_1pkt.py")
# args['output_file']     = "-"               # File to write results to, '-' for stdout
# args['warmup_mult']     = "1kpps"           # Warm up at <> multiplier (eg. "1kpps")
# args['warmup_duration'] = 1                 # Warm up for <> seconds (eg. 30)
# args['rampup_target']   = 20                # Ramp up to <> % of line rate (eg. 100)
# args['rampup_duration'] = 30                # Ramp up to target in <> seconds (eg. 180)
# args['hold_duration']   = 10                # Hold at target for <> seconds (eg. 30)
#
# Using an STL profile given in <profile_file>, this loadtest will start a
# loadtest at <warmup_mult> for <warmup_duration> seconds, after which a
# <rampup_target> (which is a percentage of line rate) is linearly achieved
# during <rampup_duration> seconds of runtime. When the target is reached, it
# is held there for <hold_duration> seconds, after which the loadtest traffic
# is stopped.
#
# Every second, stats are grabbed and added to an array. That array is written
# to a file named <output_file> when the loadtest is finished.
#
# Several invariants are held, in particular packet input/output error rates
# and loadtester CPU and queue load are measured. When they are out of bounds,
# the loadtest is aborted.

def load_test (vars):
    # create client
    c = STLClient(server = vars['server'])
    passed = True

    print("Running against %(server)s profile %(profile_file)s, warmup %(warmup_mult)s for %(warmup_duration)ds, rampup target %(rampup_target)d%% of linerate in %(rampup_duration)ds, hold for %(hold_duration)ds output goes to %(output_file)s" % vars)

    output={}
    output['vars'] = vars
    output['stats'] = {}

    try:
        # connect to server
        c.connect()

        # take all the ports
        c.reset()
        c.clear_stats()

        dir_0 = [0]
        dir_1 = [1]

        print("Mapped ports to sides {0} <--> {1}".format(dir_0, dir_1))

        # load IMIX profile
        profile_file = os.path.join(stl_path.STL_PROFILES_PATH, vars['profile_file'])
        profile1 = STLProfile.load_py(profile_file, direction=0)
        profile2 = STLProfile.load_py(profile_file, direction=1)
        #stream1 = profile1.get_streams(tunables=vars['profile_tunables'])
        #stream2 = profile2.get_streams(tunables=vars['profile_tunables'])
        stream1 = profile1.get_streams()
        stream2 = profile2.get_streams()

        # add both streams to ports
        c.add_streams(stream1, ports = dir_0)
        c.add_streams(stream2, ports = dir_1)

        # clear the stats before starting
        c.clear_stats()

        c.start(ports = (dir_0 + dir_1), mult = vars['warmup_mult'], total = True)
        print("Warming up {0} <--> {1} at rate of {2} for {3} seconds".format(dir_0, dir_1, vars['warmup_mult'], vars['warmup_duration']))
        time.sleep(vars['warmup_duration']);
        c.pause(ports = (dir_0 + dir_1))
        time.sleep(0.100);
        stats = c.get_stats()
        output['stats']['first'] = stats
        if vars['unidirectional']:
           used_ports = dir_0
        else:
           used_ports = (dir_0 + dir_1)
        c.resume(ports = used_ports)

        current_pct = target_pct = 0
        c.clear_stats()
        for i in range(vars['rampup_duration']+vars['hold_duration']):

            target_pct = int((vars['rampup_target'] * i) / vars['rampup_duration']);
            if target_pct < 1:
                target_pct = 1
            if target_pct > vars['rampup_target']:
                target_pct = vars['rampup_target']
            if (target_pct != current_pct):
                current_pct = target_pct
                my_mult = ("%d%%"%(current_pct))
                c.update(ports = used_ports, mult = my_mult)
                print("Setting load {0} <--> {1} to {2} of linerate".format(dir_0, dir_1, my_mult))

            if (i==vars['rampup_duration']):
                print("Holding load {0} <--> {1} at {2} of linerate for {3} seconds ".format(dir_0, dir_1, my_mult, vars['hold_duration']))

            time.sleep(1)
            stats = c.get_stats()
            output['stats'][i] = stats
            print("stats: %10spps %10sbps (%.2f%% of linerate)" % (fmt_xps(stats['total']['rx_pps']), fmt_xps(stats['total']['rx_bps_L1']), stats['total']['rx_util']))

            total_ierrors = stats['total']['ierrors']
            total_oerrors = stats['total']['oerrors']
            total_errors = total_ierrors + total_oerrors

            total_ipackets = stats['total']['ipackets']
            total_opackets = stats['total']['opackets']
            total_packets = total_ipackets + total_opackets

            global_cpu_util = stats['global']['cpu_util']
            global_queue_full = stats['global']['queue_full']

            if (abs(total_ipackets - total_opackets) > 0.001 * total_packets):
                print("WARNING: DUT packetloss too high")
                passed = False

            if (total_errors > 0.01 * total_packets):
                print("WARNING: Loadtester errors too high")
                passed = False

            if (global_queue_full > 0.01 * total_packets):
                print("WARNING: Loadtester saturated")
                passed = False

        # We're done - pause the traffic and wait for a while to ensure all in transit packets can arrive.
        c.pause(ports = used_ports)
        time.sleep(0.100);
        stats = c.get_stats()
        output['stats']['last'] = stats

        print("Loadtest finished, stopping")
        c.stop(ports = used_ports)

    except STLError as e:
        passed = False
        print(e)

    finally:
        c.disconnect()

    if passed:
        print("\nTest has passed :-)\n")
    else:
        print("\nTest has failed :-(\n")
    if vars['output_file'] == "-":
        pprint(output)
    else:
        print("Writing output to %s" % vars['output_file'])
        outFile=open(vars['output_file'], 'w')
        json_output = json.dumps(output, indent = 4)
        outFile.write(json_output)
        outFile.close()
      

parser = argparse.ArgumentParser(description="T-Rex Stateless Loadtester -- pim@ipng.nl")
parser.add_argument('-s', '--server',
                    dest='server',
                    help='Remote trex address (default: 127.0.0.1)',
                    default='127.0.0.1',
                    type = str)

parser.add_argument('-p', '--profile',
                    dest='profile_file',
                    help='STL profile file to replay (default: imix.py)',
                    default='imix.py',
                    type = str)

parser.add_argument('-pt', '--profile_tunables',
                    dest='profile_tunables',
                    help='STL profile tunables (default: "")',
                    default='',
                    type = str)

parser.add_argument('-o', '--output',
                    dest='output_file',
                    help='File to write results into, use "-" for stdout (default: -)',
                    default='-',
                    type = str)

parser.add_argument('-u', '--unidirectional',
                    dest='unidirectional',
                    help='Send only oneway traffic (from port0 to port1) (default: False)',
                    action='store_true')

parser.add_argument('-wm', '--warmup_mult',
                    dest='warmup_mult',
                    help='During warmup, send this "mult" (default: 1kpps)',
                    default='1kpps',
                    type = str)

parser.add_argument('-wd', '--warmup_duration',
                    dest='warmup_duration',
                    help='Duration of warmup, in seconds (default: 30)',
                    default=30,
                    type = int)

parser.add_argument('-rt', '--rampup_target',
                    dest='rampup_target',
                    help='Target percentage of line rate to ramp up to (default: 100)',
                    default=100,
                    type = int)

parser.add_argument('-rd', '--rampup_duration',
                    dest='rampup_duration',
                    help='Time to take to ramp up to target percentage of line rate, in seconds (default: 600)',
                    default=600,
                    type = int)

parser.add_argument('-hd', '--hold_duration',
                    dest='hold_duration',
                    help='Time to hold the loadtest at target percentage, in seconds (default: 30)',
                    default=30,
                    type = int)

load_test(vars(parser.parse_args()))
